import numpy

__all__ = ['eye', 'ones', 'zeros', 'kk_proj', 'f_to_vals']


# convenience helpers to ensure matrices
def eye(*args):
    return numpy.matrix(numpy.eye(*args))

def ones(*args):
    return numpy.matrix(numpy.ones(*args))

def zeros(*args):
    return numpy.matrix(numpy.zeros(*args))


#-----
def kk_proj(x, bounds):
    """
  Projection onto the feasible hyperrectangle.
 
  px = kk_proj(x, bounds)
 
  Not exciting stuff.
 
  C. T. Kelley, September 15, 2008
  This code comes with no guarantee or warranty of any kind."""

    px = numpy.minimum(bounds[:,1],  x)
    return numpy.maximum(bounds[:,0], px)


#-----
def f_to_vals(funs, least_squares):
    """
  Evaluate fvals = f^T f/2 when f is a vector least squares residual.
 
  fvals = f_to_vals(funs, least_squares)
 
  There is no reason you'd want to mess with this.
 
  C. T. Kelley, September 15, 2008
  This code comes with no guarantee or warranty of any kind."""

    try:
        n = funs.shape[0]
    except Exception:
        n = 1
        funs = numpy.array([funs])

    if least_squares == 1:
        fvals = zeros(n)
        for i in range(n):
            fvals[:,i] = funs[:,i].dot(funs[:,i])/2
    else:
        fvals = funs

    return fvals



from __future__ import print_function
# Python version of IMFIL v1.02 "imfil.m" MATLAB version by C.T. Kelly.
#
# Modified and redistributed with permission.

from SQCommon import Result, ObjectiveFunction
from ._optset import optset
from ._util import error_check, check_first_eval, isok
from ._linalg import eye, ones, zeros, kk_proj, f_to_vals
from ._history import CompleteHistory, append_history, single_point_hist_update, \
   many_point_hist_update, scan_history
from ._exec_functions import gauss_newton, qn_update
import collections
import copy
import logging
import math
import numpy
import operator
import random

__all__ = ['minimize', 'log']

log = logging.getLogger('SKQ.ImFil')


#-----
class IterData(object):
    def __init__(self, h, obounds, itc, xb, fobjb, funsb, complete_history,
                 f_internal, core_data, options):
        self.h                = h
        self.obounds          = obounds
        self.itc              = itc
        self.xb               = xb
        self.fobjb            = fobjb
        self.funsb            = funsb
        self.complete_history = complete_history
        self.f_internal       = f_internal
        self.core_data        = core_data
        self.options          = options

    def copy(self):
        return IterData(self.h, self.obounds.copy(), self.itc, self.xb.copy(),
            self.fobjb.copy(), self.funsb.copy(), self.complete_history,
            self.f_internal, self.core_data, self.options)

class StencilData(object):
    def __init__(self, stencil_delta, svarmin, noise_val, bounds, v):
        self.stencil_delta = stencil_delta
        self.svarmin       = svarmin
        self.noise_val     = noise_val
        self.bounds        = bounds
        self.v             = v


#-----
def minimize(f, x0, bounds, budget=10000, optin=None, **optkwds):
    """
  Minimization of noisy functions subject to explicit bound
  constraints + hidden constraints

  opt_x, opt_f, histout, [complete_history] = imfil(x0, f, budget, bounds, optin)

  Input:
        x0 = initial iterate.
         f = objective function.
             The calling sequence for f should be
             fout, ifail, icount = f(x,h)

              h is an optional argument, and should be used only if your
              function is scale-aware, ie does something useful with
              the scale, such as tuning accuracy.

              fout = f(x)

              fail = 0 unless the evaluation of f fails, in which case
                     set ifail=1 and fout = NaN

              icount = number of calls to the expensive part of f
                      f should be smart enough to not do anything
                      if a bound constraint or a cheap hidden constraint
                      is violated

              f may return a fourth argument noise_level if f is
              noise aware. RTFM about this one.

    budget = max cost, uses icount from f(x)
             The iteration will terminate after the iteration that
             exhausts the budget. This argument is required.
             Suggestion: try 10*N^2.

  bounds   = [low, high] = N x 2 array of bound constraints on the variables.
             This is REQUIRED. If you want to solve an unconstrained
             problem, you must do so without the scaling we do in
             f_internal, and using imfil_core is currently the only way
             to do that. The lower bounds are the first column;
             the upper bounds are the second.

             I plan let you solve unconstrained problems by setting
             the bounds to Inf or -Inf. That will take some time while
             I figure out how I want to do the scaling (or make you
             do it).

   optin   = options structure
             Documentation on the way. In the meantime, RTFM or
             look at _optset.py.

 extra_arg = optional extra argument to be passed to f

  Output:
         x = estimated minimizer

   histout = iteration history, updated after each nonlinear iteration
           = (number of iterations )x(N+5) array, the columns are
              [fcount, fval, norm(sgrad), norm(step), iarm, xval]
              fcount      = cumulative function evals
              fval        = current function value
              norm(sgrad) = current projected stencil grad norm
              norm(step)  = norm of last step
              iarm        = line searches within current iteration
                          = -1 means first iterate at a new scale
              xval        = transpose of the current iteration

  complete_history = complete evaluation history

  complete_history is a structure with the data on every evaluation of f.

  complete_history.good_points has the successful points for columns.
  complete_history.good_values is a $M x N$ matrix  of the values at
               the good points. M > 1 for least squares.
               good_values = f(good_points)

  complete_history.failed_points has the unsuccessful points for columns.

  You may want to use this to build surrogates, decide to add new points to
  the stencil, or for troubleshooting. The complete history can take up a
  lot of room. I will only return it as output if you ask for it, and will
  not store it at all if you set the complete_history option to '0'.
  If I don't store it at all, then your functions to add directions to the
  stencil can't use that data.

  C. T. Kelley, July 19, 2010.
  This code comes with no guarantee or warranty of any kind."""

    global imfil_fscale

    log.debug('optimization start')

    # TODO: get rid of matrix use
    # Make sure x0 is a single-column matrix
    if not isinstance(x0, numpy.matrix):
        x0 = numpy.matrix(x0).T

    # Make sure bounds are a 2-column martrix
    if not isinstance(bounds, numpy.matrix):
        bounds = numpy.matrix(bounds)

    qbounds = bounds
    dbounds = bounds[:,1] - bounds[:,0]

    # Objective function should be callable
    if not callable(f):
        raise ValueError('"%s" is not callable' % (str(objfunc),))

    # Make sure x0 is a single-column matrix
    if not isinstance(x0, numpy.matrix):
        x0 = numpy.matrix(x0).T

    # Use optset to fill in defaults where no option values are provided
    if type(optin) == dict:
        options = optset(**dict(optin, **optkwds))
    else:
        options = optset(optin, **optkwds)

    objfunc = ObjectiveFunction(f, options)

    n = len(x0)

    # If the smooth_problem option is on, fix the dependencies and
    # change the scales.
    if options.smooth_problem == 1:
        log.info('smooth problem selected: adjusting options');
        bscales=[.5, .01, .001, .0001, .00001]
        options = optset(options,
              custom_scales=bscales,
              stencil_wins='yes',
              limit_quasi_newton='off',
              armijo_reduction=0.25,
              maxitarm=5)

    # Propagate 'verbose' option
    if options.verbose:
        log.setLevel(logging.INFO)

    # Load the fun_data structure so f_internal will know about the scaling.
    imfil_fscale = options.fscale
    FunData = collections.namedtuple('FunData', 'objfunc qbounds dbounds fun_fscale')
    fun_data = FunData(objfunc=objfunc,
        qbounds=qbounds, dbounds=dbounds, fun_fscale=imfil_fscale)
    CoreData = collections.namedtuple('CoreData', 'options fun_data')
    core_data = CoreData(options=options, fun_data=fun_data);

    # Scale the initial iterate to [0,1] before shipping to imfil_core.
    # imfil_core uses 0 and 1 as the bounds for optimization.
    error_check('bounds', x0, bounds)

    z0 = (x0 - qbounds[:,0])/dbounds
    z, fval, histout, complete_history = \
       imfil_core(z0, f_internal, budget, core_data, bounds)

    # Unscale the results to original bounds.
    opt_par = numpy.multiply(dbounds, z) + qbounds[:,0]
    db = numpy.diagflat(dbounds)
    qb = numpy.diagflat(qbounds[:,0])
    histout = numpy.matrix(histout)
    xout = histout[:,5:n+5]
    xlow = ones(xout.shape)*qb
    histout[:,5:n+5] = xout*db + xlow
    xout = histout[:,5:n+5]

    # Correct scaling for nonlinear least squares problems.
    if not options.least_squares:
        val_scale = imfil_fscale
    else:
        val_scale = math.sqrt(imfil_fscale)

    # Unscale the complete history
    if options.standalone and options.complete_history:
        ng, mg = len(complete_history.good_points), 2
        good_points = zeros((mg, ng))
        for i in range(ng):
            good_points[0, i] = complete_history.good_points[i][0]
            good_points[1, i] = complete_history.good_points[i][1]

        complete_history.good_values = [v*val_scale for v in complete_history.good_values]
        if mg > 0:
            clow = qb*ones((mg, ng))
            complete_history.good_points = db * good_points + clow

        nf, mf = len(complete_history.failed_points), 2
        failed_points = zeros((mf, nf))
        for i in range(nf):
            failed_points[0, i] = complete_history.failed_points[i][0]
            failed_points[1, i] = complete_history.failed_points[i][1]
        if mf > 0:
            flow = qb * ones((mf, nf))
            complete_history.failed_points = db * failed_points + flow

    # Unscale the function and gradients.
    histout[:,1] = histout[:,1]*imfil_fscale
    histout[:,2] = histout[:,2]*imfil_fscale

    # Turn opt_par into an array (are a matrix) and scale minimum.
    result = Result(fval*val_scale, numpy.squeeze(numpy.asarray(opt_par)))

    if options.standalone:
        if options.complete_history:
            return result, histout, complete_history
        return result, histout
    return result, objfunc.get_history()


#-----
def imfil_core(x0, f, budget, core_data, bounds):
    """
  Minimization of f(x) subject to explicit bound constraints

  x, fval, histout, complete_history = \
          imfil_core(x0, f, budget, core_data, bounds)

  Bound constrained, parallel, implicit filtering code.

  IMPLICIT FILTERING with SR1 and BFGS quasi-Newton methods

  Input:
        x0 = initial iterate
         f = objective function,

             the calling sequence for f should be
             [fout,ifail,icount]=f(x,h)

             h is an optional argument for f, and should be used only if your
             function is scale-aware, ie does something useful with
             the scale, such as tuning accuracy.

   budget = upper limit on function evaluations. The optimization
                  will terminate soon after the function evaluation counter
                  exceeds the budget.

   core_data = structure with the options + fun_data
               The options are documented in optset.m.
               fun_data is private to imfil.m and is used in the internal
               scaling for the function evaluation.

   bounds  = N x 2 array of bound constraints on the variables.
             These are the original bounds for the problem. We only
             use these if you use the add_new_directions option.

  Output:
         x = estimated minimizer
      fval = corresponding minimum
   histout = iteration history, updated after each nonlinear iteration
           = N+five column array, the rows are
             [fcount, fval, norm(sgrad), norm(step), iarm, xval]
             fcount = cumulative function evals
             fval = current function value
             norm(sgrad) = current (projected) simplex grad norm
                         = -1 means no gradient for this xval, this
                  can happen, for example, if you hit the target before you
                  evaluate the simplex derivative
             norm(step) = norm of last step
                        = 0 means no change, eg for stencil failure
             iarm=line searches in current iteration to move to new point
                 =-1 means first iterate at a new scale or that the
                     inner iteration was terminated before the line search,
                     eg for stencil failure
              xval = transpose of the current iteration

  complete_history = complete evaluation history

  complete_history is a structure with the data on every evaluation of f.
  complete_history.good_points has the successful points for columns.
  complete_history.good_values is a $M x N$ matrix  of the values at
               the good points. M > 1 for least squares.
               good_values = f(good_points)

  complete_history.failed_points has the unsuccessful points for columns.

  You may want to use this to build surrogates, decide to add new points to
  the stencil, or for troubleshooting. The complete history can take up a
  lot of room. I will only return it as output if you ask for it, and will
  not store it at all if you set the complete_history option to '0'.
  If I don't store it at all, then your functions to add directions to the
  stencil can't use that data.

  Raises exception in case of failure.

  C. T. Kelley, July 16, 2010
  This code comes with no guarantee or warranty of any kind."""

    global imfil_fscale

    fcount = 0
    options = core_data.options
    # set up the difference scales and options
    n = len(x0)

    # In imfil_core the bounds are 0 and 1 on all the variables.
    # We use the real bounds only if you use the add_new_directions
    # option.
    obounds = zeros((n, 2)); obounds[:,1] = 1

    # Get the options we need.
    imfil_maxit    = options.maxit
    imfil_target   = options.target
    imfil_termtol  = options.termtol
    stencil_wins   = options.stencil_wins
    complete_history = CompleteHistory()
    assert not complete_history.good_values

    # The explore option may require some setup.
    explore_function, explore_data_flag, explore_data = setup_explore(options)

    # Initialize the iteration; create the stencil; set up the scales; ...
    x = x0.copy(); xold = x0.copy(); histout = []
    dscal = create_scales(options); nscal = len(dscal)
    if options.executive == 1:
        hess = options.executive_data
    else:
        hess = eye(n)

    xc = x0.copy(); ns = 0; failc = 0
    stop_now = False
    fval = imfil_target+1

    # Sweep through the scales.
    sflag = 1
    while ns < nscal and fcount <= budget and failc < options.maxfail and \
           not stop_now and fval > imfil_target:
        h = dscal[ns]
        log.debug('current scale: %f', h)

        # Evaluate the function to test for instant termination and (if
        # noise_aware is on) to get the estimate of the noise. Both the noise
        # and the value of f  may vary as a function of the scale, so we
        # have to test it every time.
        #
        # We could move this outside of the main loop if noise_aware and scale_aware
        # are both off. In any case, we don't have to reevaluate f after a stencil
        # failure unless f is scale-aware.
        if options.noise_aware > 0 or fcount == 0 or options.scale_aware > 0:
            funs, iff, icf, noise_val = f(x, h, core_data)
            fval = f_to_vals(funs, options.least_squares)
            icount = icf
        else:
            icount = 0

        # The first call to f is the place to check for sanity.
        #
        # The first call to f also sorts out the scaling. So you can't scale the
        # the targets, errors, or deltas before that call.
        if fcount == 0:
            error_check('first_eval', funs, options, iff)

            imfil_target = imfil_target/imfil_fscale
            function_delta = options.function_delta/imfil_fscale
            append_history(histout, 1, fval, 0, 0, 0, x)

            # Initialize the internal data structures.
            itc = 0
            stencil_data = \
                create_stencil_data(options, imfil_fscale, noise_val, bounds)
            iteration_data = IterData(h=h, obounds=obounds, itc=itc,\
               xb=x, fobjb=fval, funsb=funs, complete_history=complete_history,\
               f_internal=f, core_data=core_data, options=options)
        else:
            # End of first-call-to-f block
            #
            # If it's not the first call to f, update the internal data structures.
            iteration_data.h = h
            stencil_data.noise_val = noise_val;

        if options.complete_history:
            single_point_hist_update(iteration_data.complete_history, x, funs, iff)

        fcount += icf
        if fval < imfil_target :
            stop_now = True
            break

        stol = imfil_termtol*h; iarm=0; nfail=0

        # Compute the stencil gradient to prepare for the quasi-Newton iteration.
        sdiff, sgrad, npgrad, fcount, sflag, jac, iteration_data, stop_now = \
            manage_stencil_diff(x, f, funs, \
                  iteration_data, fcount, stencil_data)
        append_history(histout, fcount, fval, npgrad, -1, -1, x)
        gc = sgrad

        #
        if npgrad < stol or sflag == 0 or stop_now:
            # Declare convergence at this scale on stencil failure or tolerance match.
            if npgrad < stol:
                log.debug('converged at current scale with gradient (%f) below cut-off (%f)',
                          npgrad, stol)
            else:
                log.debug('ended current scale due to stencil failure')
            gc = sgrad
            if sflag != 0:
                failc += 1
            else:
                failc = 0
        else:
            # Take a few quasi-Newton iterates. This is the inner iteration.
            log.debug('start inner Newton loop iteration')
            failc = 0

            # itc = inner iteration counter
            itc = 0

            # Newton while loop
            while itc < imfil_maxit*n and fval > imfil_target and \
                  npgrad >= stol and nfail==0 and fcount < budget and sflag > 0:

                itc += 1
                iteration_data.itc = itc;
                fc = fval
                # Take an inner iteration. sdiff is the simplex derivative
                # (gradient or Jacobian) at x. gc is the simplex gradient at xc.
                xp, fval, funs, fcount, iarm, iteration_data, nfail, hess = \
                    inner_iteration(f, x, funs, sdiff, xc, gc, \
                                    iteration_data, hess, fcount)

                # Stop the entire iteration if you've hit the target.
                if fval < imfil_target:
                    stop_now = True
                    x = xp
                    stepn = numpy.linalg.norm(xold-x, ord=numpy.inf)
                    append_history(histout, fcount, fval, npgrad, stepn, iarm, x)
                    break

                # Update xold, xc, and x. At this point the model Hessian and gc
                # are evaluated at xc. xc and gc only get updated right here.
                xold = x.copy(); xc = x.copy(); gc = sgrad; x = xp.copy()

                # If stencil_wins is on, then take the best point you have.
                # You'll update x, but not xc.
                if stencil_wins == 1:
                    x, funs, fval = write_best_to_x(iteration_data)
                    stepn = numpy.linalg.norm(xc-x, ord=numpy.inf)
                    append_history(histout, fcount, fval, npgrad, stepn, iarm, x)
                    nfail = 0

                # Stop on small objective function changes?
                fdelta = abs(fval-fc)
                if fdelta > 0 and fdelta < function_delta :
                    stop_now = True
                    sflag = 0
                    stepn = numpy.linalg.norm(xc-x, ord=numpy.inf)
                    append_history(histout, fcount, fval, npgrad, stepn, iarm, x)
                    break

                #    Compute the difference gradient for the next nonlinear iteration.
                if not stop_now and fcount < budget:
                    sdiff, sgrad, npgrad, fcount, sflag, jac, iteration_data, stop_now = \
                        manage_stencil_diff(x, f, funs,
                            iteration_data, fcount, stencil_data)

                # If the quasi-Newton method terminated successfully or with a
                # stencil failure, make sure x is now the best point.
                if stop_now or sflag == 0:
                    iteration_data, rflag = \
                        reconcile_best_point(funs, xp, iteration_data);
                    x, funs, fval = write_best_to_x(iteration_data)
                    stepn = numpy.linalg.norm(xc-x, ord=numpy.inf)
                    append_history(histout, fcount, fval, npgrad, stepn, -1, x)
                    break

                # Otherwise, update the history array and keep going.
                else:
                    stepn = numpy.linalg.norm(xc-x, ord=numpy.inf)
                    append_history(histout, fcount, fval, npgrad, stepn, iarm, x)

                # Update xold. It will not be the same as xc (the point for the most
                # recent stencil derivative point) if you are exiting the quasi-Newton
                # loop.
                xold = x.copy()

                # end Newton while loop

            log.debug('end inner Newton loop iteration')
        # end test for stencil failure for first derivative at new scale

        # Apply the explore_function if there is one.
        if options.explore == 1:
            fcount, iteration_data = \
                manage_explore(explore_function, f, iteration_data, fcount, explore_data)

        # After the quasi-Newton iterate, I make sure that the
        # quasi-Newton point is the best I've seen. If it's not, I fix it.
        # So, 'stencil_wins' is half-way on. One consequence of this is that
        # I take the best point if the inner iteration fails.
        iteration_data, rflag = reconcile_best_point(funs, x, iteration_data)
        x, funs, fval = write_best_to_x(iteration_data)

        if rflag == 1:
            if histout[-1][0] == fcount:
                histout = histout[:-1]

        append_history(histout, fcount, fval, npgrad, 0, 0, x)

        if options.verbose:
            log.info(ns, fval, h, npgrad, itc, fcount)

        ns += 1
        # end of loop over scales

    return x, fval[0], histout, complete_history


#-----
def f_internal(x, h, core_data):
    """
  Creates an O(1) dummy function with unit bounds.
  So, (I hope) the gradient and stencil are reasonably scaled.

  I encode scale and noise awareness in this function, too. If your
  function is not noise aware, I pass a zero noise back to imfil_core.
  imfil_core will do the right thing after that."""

    global imfil_fscale

    options = core_data.options
    exarg = options.extra_arg_value
    fun_data = core_data.fun_data
    objfunc = fun_data.objfunc
    qbounds = fun_data.qbounds
    dbounds = fun_data.dbounds

    mx, nx = x.shape
    z = x.copy()
    for ix in range(nx):
        z[:,ix] = numpy.multiply(dbounds, x[:,ix])+qbounds[:,0]
    zargs = numpy.squeeze(numpy.asarray(z.T))

    call_args = [zargs,]
    if options.scale_aware:
        call_args.append(h)
    if options.extra_argument:
        call_args.append(exarg)

    res = objfunc(*call_args)
    fx, iff, icf, tol = 0., 0, 1, 0.
    if type(res) == tuple:
        fx = res[0]
        if 1 < len(res):
            iff = res[1]
        if 2 < len(res):
            icf = res[2]
        if 3 < len(res):
            tol = res[3]
    else:
        fx = res
        mz, nz = z.shape
        if nz != 1:
           iff = zeros(nz); icf = nz*ones(nz)

    # If this is the first time you evalute f and if imfil_fscale < 0,
    # we change imfil_fscale to the correct relative scaling factor. This
    # also makes imfil_fscale > 0 so we only compute the scaling once.
    if imfil_fscale <= 0:
        if imfil_fscale == 0:
            imfil_fscale = -1.2

        if options.least_squares:
            val = fx[:,0].dot(fx[:,0])/2.
            scale_base = val
        else:
            scale_base = abs(fx)

        if scale_base == 0:
            imfil_fscale = 1
        else:
            imfil_fscale = abs(imfil_fscale)*scale_base

    # Scale the function value.
    if not options.least_squares:
        fx /= imfil_fscale;
    else:
        fx /= math.sqrt(imfil_fscale)
    tol /= imfil_fscale

    return fx, iff, icf, tol


#-----
def setup_explore(options):
    """
  Gets the explore function organized if you have one."""

    explore_function = []
    explore_data_flag = 0
    explore_data = []
    if options.explore:
        explore_function  = options.explore_function
        explore_data_flag = options.explore_data_flag
    if explore_data_flag:
        explore_data = options.explore_data

    return explore_function, explore_data_flag, explore_data


#-----
def create_scales(options):
    """
  Creates the scales for imfil.m. Nothing much here right now, but
  custom scales ... are in the near future.

  dscal = create_scales(options)

  C. T. Kelley, September 15, 2008
  This code comes with no guarantee or warranty of any kind."""

    custom_scales = options.custom_scales
    mcs = len(custom_scales)
    if mcs > 0:
        dscal = custom_scales
    else:
        step  = options.scale_step
        start = options.scale_start
        depth = options.scale_depth

        # Do some error checking. Warn or complain as needed.
        if start > depth:
            raise ValueError('imfil_create_scales: error in scales, start > depth')

        # scale reduces with power of 2
        dscal = [step**-x for x in range(start, depth+1)]

    return dscal


#-----
def create_stencil_data(options, fscale, noise_val, bounds):
    """
  Build a structure with the data, targets, errors, and deltas that
  you need to compute the stencil derivative and  determine stencil failure.
  This structure gets passed to manage_stencil_diff and contains everything
  that depends only on the scale and the options. I'm doing this mostly
  to keep the argument list from occupying several lines and confusing me.

  stencil_data = create_stencil_sdata(options, imfil_fscale, noise_val, bounds)

  Do not attempt to land here."""

    n = len(bounds[:,1])
    v = create_stencil(options, n)
    stencil_delta = options.stencil_delta/fscale
    svarmin       = options.svarmin/fscale

    stencil_data = StencilData(
        stencil_delta=stencil_delta, svarmin=svarmin,
        noise_val=noise_val, bounds=bounds, v=v)
    return stencil_data


#-----
def create_stencil(options, n):
    """
  Builds the stencil for imfil.m. As we evolve this we will be
  adding the ability to do random rotations for all or part of
  a stencil and all sorts of other stuff.

  vstencil = create_stencil(options, n)

  Input:
         options = imfil.m options structure
               n = dimension

  Output:
               v = stencil

  C. T. Kelley, July 14, 2009
  This code comes with no guarantee or warranty of any kind."""

    stencil  = options.stencil
    vstencil = options.vstencil

    # Is vstencil really there?
    if vstencil:
        mv, nv = vstencil.shape
        if mv+nv > 0:
            stencil = -1

    # Build the stencil.
    if stencil == -1:    # Custom stencil is ok.
        pass
    elif stencil == 0:   # central difference stencil
        v = numpy.hstack((numpy.eye(n), -numpy.eye(n)))
    elif stencil == 1:   # one sided stencil = ON THE WAY OUT!
        v = numpy.eye(n)
    elif stencil == 2:   # positive basis stencil
        v = numpy.hstack((numpy.eye(n), -numpy.ones(n)/math.sqrt(n)))
    else:
        raise ValueError('illegal stencil in _imfil.create_stencil')

    # If vstencil is a custom job, take it. Otherwise use one of
    # the internal choices.
    if stencil != -1:
        vstencil = v

    return numpy.matrix(vstencil)


#-----
def stencil_diff(x, f, dx, fc, iteration_data, complete_history):
    """
  Stencil derivative for optimization and nonlinear least squares.
  This function also tests for best point in stencil.

  grad, best_value, best_value_f, best_point, icount, sflag,svar, diff_hist, jac = \
      stencil_diff(x, f, dx, fc, iteration_data, options, h, complete_history)

    Input:  x  = current point
            f  = objective function (or vector residual for
                 nonlinear least squares)
            dx = scaled difference directions
            fc = current function value

    iteration_data = structure with the options + fun_data +
                     current iteration parameters.
                The options are documented in optset.m.
                fun_data is private to imfil.m and is used in the internal
                scaling for the function evaluation.

  complete_history = every point imfil has ever seen. That data are used
                     here to prevent redundant evaluations.

    Output:
          grad = stencil gradient; NaN if every point on the stencil is bad
   best_value  = best value in stencil
   best_value_f= best least squares residual in stencil
   best_point  = best point in stencil
        icount = counter of calls to expensive part of function
                  the function needs to return this.
         sflag = 0 current point is best in the stencil
                      This means stencil failure.

               = 1 means there's a better point on the stencil.

          svar = variation on stencil = max - min

     diff_hist = every function evaluation for points satisfying the
                  bounds stored in a nice struct. This is used to update
                  the complete_history structure.

           jac = stencil Jacobian for least squares problems.

  C. T. Kelley, July 30, 2009
  This code comes with no guarantee or warranty of any kind."""

    core_data = iteration_data.core_data
    h = iteration_data.h

    # bounds = iteration_data.obounds are the 0-1 bounds for the
    # scaled problem.
    bounds = iteration_data.obounds

    options = core_data.options
    least_squares = options.least_squares

    # Poll the stencil and collect some data.
    best_value, best_value_f, best_point, icount, sgood, good_points, good_values,\
         good_dx, good_df, failed_points = \
           poll_stencil(x, f, dx, fc, bounds, core_data, h, complete_history)

    # Initialize the search for best value at the center.
    if least_squares:
        fval = fc.dot(fc)/2   # best value in the stencil
    else:
        fval = fc

    # Use the points and values to get the stencil statisitcs.
    sflag, best_value, best_value_f, best_point, svar, diff_hist = \
        collect_stencil_data(good_points, good_values, failed_points,
           x, fval, fc, fc, options)

    # sgood = 0 means there are no good points. Shrink time!
    #
    # Estimate the derivative.
    # The idea is that fprime*dx approx df, where
    # fprime   = m x n is the gradient-transpose or the Jacobian
    # good_dx  = n x pnew (pnew <= vsize) is the matrix of good steps
    # good_df  = m x pnew is the collection of good function values at x + dx
    #
    # So the least squares problem to be solved is
    #         min || fprime * good_dx - good_df ||
    # for the columns of fprime. The minimum norm solution is
    # fprime = good_df*pinv(good_dx).
    if sgood > 0:
        pdx = numpy.linalg.pinv(good_dx)
        fprime = good_df.dot(pdx)
        if least_squares == 0:
            grad = fprime.T
            jac=[]
        else:
            jac = fprime
            grad = fprime.dot(fc)
    else:
        grad = numpy.NaN*x
        jac = []

    return grad, best_value, best_value_f, best_point, icount, sflag, svar, diff_hist, jac


#-----
def manage_stencil_diff(x, f, funs, iteration_data, fcount, stencil_data):
    """
  This function calls stencil_diff to compute the stencil derivative,
  updates the evaluation counter and complete_history, runs through
  all the tests for stencil failure, and sends back everything
  imfil_core needs to do its job.

  This is an internal function. There is no reason to hack this code.
  Don't do it.

  This function is under development and changes frequently.

  C. T. Kelley, January 12, 2011"""

    options = iteration_data.options

    h = iteration_data.h

    # Unpack the stencil_data structure.
    stencil_delta = stencil_data.stencil_delta
    svarmin       = stencil_data.svarmin
    noise_val     = stencil_data.noise_val
    obounds       = iteration_data.obounds
    bounds        = stencil_data.bounds
    v             = stencil_data.v

    # Complete the direction matrix and compute the stencil derivative.
    vv = augment_directions(x, v, h, options, bounds)
    complete_history = iteration_data.complete_history
    sgrad, fb, fbf, xb, icount, sflag, svar, diff_hist, jac = \
        stencil_diff(x, f, h*vv, funs, iteration_data, complete_history)
    fcount += icount
    pgrad = x - kk_proj(x-sgrad, obounds)
    npgrad = numpy.linalg.norm(pgrad, ord=numpy.inf)

    # Run the optional stencil failure tests.
    #
    # If noise_aware = 1 (i.e. noise_val returned) and the scaled variation
    # in f is < than the function's estimate of the noise, then I declare
    # stencil failure!
    if max(noise_val, svarmin) > svar:
        log.debug('scaled variation (%f) is less than estimated noise (%f)', svar, noise_val)
        sflag = 0

    # If stencil_delta > 0, then I terminate the entire optimization
    # when the scaled variation in f < stencil_delta. I report this as
    # stencil failure as well.
    stop_now = False
    if stencil_delta > svar:
        log.debug('scaled variation (%f) is less than stencil delta (%f)', svar, stencil_delta)
        stop_now = True
        sflag = 0

    # Update the iteration_data structure.
    iteration_datap = iteration_data
    if options.complete_history:
        complete_history = many_point_hist_update(complete_history, diff_hist, False)
        iteration_datap.complete_history = complete_history

    iteration_datap, rflag = reconcile_best_point(fbf, xb, iteration_datap);

    sdiff = jac_or_grad(sgrad, jac, options)

    return sdiff, sgrad, npgrad, fcount, sflag, jac, iteration_datap, stop_now


#-----
def poll_stencil(x, f, dx, fc, bounds, core_data, h, complete_history):
    """
  Poll the stencil and organize the results.
  I do not think you will want to modify this code.

  best_value, best_point, icount, sgood, good_points, good_values, \
          good_dx, good_df, failed_points = \
      poll_stencil(x, f, dx, fc, bounds, core_data, h, complete_history)

  Input:
           x = center of stencil
           f = objective function
          dx = h*V = array of scaled directions
          fc = f(x)
      bounds = N x 2 array of lower/upper bounds
   core_data = imfil core_data structure
           h = current scale

  Output:

  The best_point data is frozen at xc for this function. We only keep
  track of it here so we can ship it to
      best_point    = f(best_point) == best_value
      best_value    = lowest value of f on the stencil
      best_value_f  = best least squares residual
      icount        = cost in units of calls to f
      sgood         = number of successful evaluations
      good_points   = list of points at which f returned a value
      good_values   = f(good_points)
      good dx       = vector of good_points - x
      good df       = vector of f(good_points) - fc
      failed_points = list of points at which f did not return a value

  C. T. Kelley, Aug 13, 2009
  This code comes with no guarantee or warranty of any kind."""

    options = core_data.options
    least_squares = options.least_squares

    n, vsize = dx.shape

    # Record the best point and best function value in the stencil.
    # Get started by using the center point.
    best_point = x
    if least_squares:
        best_value = fc.dot(fc)/2  # best value in the stencil
    else:
        best_value = fc
    best_value_f = fc

    iflag = zeros((vsize, 1))

    try:
        m = len(fc)
    except TypeError:
        m = 1
    fp = zeros((m, vsize))

    failed_points = []
    good_points = []
    good_values = []
    sgood = 0
    fval = zeros(vsize)
    icount = 0

    # fp[:,i] and fval(i) are not defined outside of the bounds
    # or if iflag(i) == 1.
    #
    # First cull the points which violate the bound constraints.
    #
    # Collect the feasible points, differences, and functions in xp1 and dx1.
    pold = 0
    dx1 = None
    xp1 = None
    xp  = zeros(dx.shape)

    # One-sided differences may need to flip the direction if the positive
    # perturbation is not feasible.
    if options.stencil == 1:
        for i in range(vsize):
            xp[:,i] = x+dx[:,i]
            if isok(xp[:,i], bounds) == 0:
                dx[:,i] = -dx[:,i]

    for i in range(vsize):
        xp[:,i] = x+dx[:,i]
        if isok(xp[:,i], bounds):
            pold = pold+1
            if dx1 is None:
                dx1 = dx[:,i]
                xp1 = xp[:,i]
            else:
                dx1 = numpy.hstack((dx1, dx[:,i]))
                xp1 = numpy.hstack((xp1, xp[:,i]))

    try:
        fp = fp[:,0:pold]
    except TypeErrorError:
        pass

    # Query the complete_history structure to see if you've evaluated f
    # at any of these points before.
    oldresults, newpoints = scan_history(complete_history, xp1, fp, dx1)

    # Copy over pre-existing values.
    iflago = zeros((vsize, 1))
    for idx, vals in oldresults.items():
        fp[:,idx] = vals[0]
        iflago[idx] = vals[1]

    # Evaluate f, in parallel if possible. Flag the failed points.
    fp1 = []
    iflag = []
    if not options.parallel:
        for point in newpoints.values():
            fpx, iflagx, ict, tol = f(point, h, core_data)
            fp1.append(fpx)
            iflag.append(iflagx)
            icount += ict
    elif newpoints:
        fp1, iflag, ictrp, tol = f(numpy.hstack(newpoints.values()), h, core_data)
        icount += sum(ictrp)

    fp = numpy.zeros((1, len(fp1)))

    # Copy over new values.
    if newpoints:
        for i, j in enumerate(newpoints.keys()):
            fp[:,j]  = fp1[i]
            iflago[j] = iflag[i]

    fp1 = fp.T; iflag = iflago

    # Identify the failed points.
    ibad = numpy.array((iflag[0:pold] == 1))[:,0]
    if ibad.any():
        failed_points = xp1[:,ibad]

    # Store the good points and their function values.
    igood = numpy.array((iflag[0:pold] == 0))[:,0]
    sgood = igood.sum()
    good_df = zeros(fp1.shape)
    good_dx = []
    if sgood > 0:
        good_points = xp1[:,igood]
        if least_squares == 1:
            good_fp = fp1[:,igood]
        else:
            good_fp = fp1[igood]
        good_dx = dx1[:,igood]
        for ig in range(sgood):
            if least_squares == 1:
                good_df[:,ig] = good_fp[:,ig] - fc
            else:
                good_df[ig] = good_fp[ig] - fc
        good_values = good_fp

    return (best_value, best_value_f, best_point, icount, sgood, \
        good_points, good_values, good_dx, good_df.T, failed_points)


#-----
def collect_stencil_data(good_points, good_values, failed_points,
        best_point_old, best_value_old, best_value_f_old, fc, options):

    least_squares = options.least_squares

    # Who's number one?
    good_scalars = f_to_vals(good_values, least_squares)
    ibest, xbest_value = min(enumerate(good_scalars), key=operator.itemgetter(1))
    xbest_point = good_points[:,ibest]
    if least_squares:
        xbest_value_f = good_values[:,ibest]
    else:
        xbest_value_f = good_values[ibest]
    if xbest_value < best_value_old:
        best_value = xbest_value
        best_value_f = xbest_value_f
        best_point = xbest_point
    else:
        best_value = best_value_old
        best_value_f = best_value_f_old
        best_point = best_point_old

    # Find the big loser.
    worst_value = max(good_scalars)

    # Assemble the history structure.
    DiffHist = collections.namedtuple('DiffHist', 'good_points good_values failed_points')
    diff_hist = DiffHist(good_points=good_points,
                    good_values=good_values, failed_points=failed_points)

    # What's the total variation?
    svar = worst_value-best_value

    # Stencil failure?
    sflag = 1
    if abs(best_value_old-best_value) < 1.E-14:
        global imfil_fscale
        log.debug('no improvement found on stencil (current best: %f)',
                  imfil_fscale*best_value_old)
        sflag = 0
        jac = []
        grad = []

    return sflag, best_value, best_value_f, best_point, svar, diff_hist


#-----
def reconcile_best_point(funs, x, old_data):
    """
  new_data, rflag = reconcile_best_point(funs, x, old_data)

  After the poll, or when it's time to terminate the inner or outer
  iteration, you may have a new best point.
  This function updates the record of the best point using the
  iteration_data structure.

  Input: x        = current point
         funs     = f(x)
     old_data     = iteration_data structure
   least_squares  = Are we solving a nonlinear least squares problem?

  Output: new_data     = updated iteration_data structure
          fvalout = f(xout)
          rflag = 0 if xout = x;  (ie f(x) is best, new best point)
          rflag = 1 if xout = xb; (best point unchanged)"""

    new_data = old_data.copy()

    rflag = 1
    fb = old_data.fobjb
    fval = f_to_vals(funs, old_data.options.least_squares)
    if fval < fb:
        rflag = 0
        new_data.xb    = x
        new_data.funsb = funs
        new_data.fobjb = fval

    return new_data, rflag


#-----
def inner_iteration(f, x, fx, sdiff, xc, gc, iteration_data, hess, fcount):
    """
  Manage the various inner iteration options.

  Inputs:
          f = objective function; f returns an residual vector.
          x = current point.
        fun = current (vector) residual at x.
        jac = stencil Jacobian at x.
         xc = previous point; xc is not used in this function, but
              that may change if we elect to put a quasi-Newton
              method in here. xc=x for the first iteration in the
              inner iteration.
         gc = stencil gradient at xc.
  iteration_data = internal structure with many goodies inside; rtfm
    hessold = Gauss-Newton model Hessian at xc. Dummy argument
              waiting for quasi-Newton or Levenberg-Marquardt.

  Output:
       hess = Gauss-Newton model Hessian at x; no update done in here.
              It's in the argument lists only to enable a quasi-Newton
              update or a Levenberg-Marquardt iteration.
         xp = new point.
      fvalp = least squares error at xp = funs'*funs/2;
       funs = vector residual at xp.
       qfct = cost in function evaluations.
       iarm = number of step size/trust region radius reductions.
  diff_hist = history data for the Gauss-Newton loop
      nfail = 0 if the line search/trust region/LM succeeds, 1 if it fails"""

    options          = iteration_data.options
    core_data        = iteration_data.core_data
    least_squares    = options.least_squares
    obounds          = iteration_data.obounds
    opt_exec         = options.executive
    if opt_exec:
        opt_exec_function = options.executive_function

    if not opt_exec:
        if not least_squares:
            xp, fval, funs, fct, iarm, diff_hist, nfail, hess = \
                qn_update(f, x, fx, sdiff, xc, gc, iteration_data, hess)
        else:
            xp, fval, funs, fct, iarm, diff_hist, nfail, hess = \
                gauss_newton(f, x, fx, sdiff, xc, gc, iteration_data, hess)
    else:
        xp, fval, funs, fct, iarm, diff_hist, nfail, hess = \
            opt_exec_function(f, x, fx, sdiff, xc, gc, iteration_data, hess)

    fcount += fct
    iteration_datap = iteration_data
    if options.complete_history:
        many_point_hist_update(iteration_data.complete_history, diff_hist)

    iteration_datap, rflag = reconcile_best_point(funs, xp, iteration_datap)
    return (xp, fval, funs, fcount, iarm, iteration_datap, nfail, hess)


#-----
def write_best_to_x(iteration_data):
    """
  Makes the best point the current iterate.

  x, funs, fval = write_best_to_x(iteration_data)"""

    x    = iteration_data.xb
    funs = iteration_data.funsb
    fval = iteration_data.fobjb
    return x, funs, fval


#-----
def augment_directions(x, vin, h, options, bounds):
    """
  Enrich the stencil with random directions and a user-supplied
  new_direcions function.

  C. T. Kelley, August 13, 2009.
  This code comes with no guarantee or warranty of any kind.

  Add the random directions."""

    vout = random_augment(vin, options.random_stencil)

    # See if there's a new_directions function.
    new_directions = options.add_new_directions
    if new_directions:
        dbv = bounds[:,1] - bounds[:,0]; db = numpy.diagflat(dbv)
        unscaled_x = db*x + bounds[:,0]
        unscaled_v = db*vout
        unscaled_vnew = new_directions(unscaled_x, h, unscaled_v)
        mv, nv = unscaled_vnew.shape
        if mv > 0:
            vnew = numpy.linalg.inv(db)*unscaled_vnew;
        mv, nv = vnew.shape
        for i in range(nv):
            vnew[:,i] = vnew[:,i]/numpy.linalg.norm(vnew[:,i], ord=numpy.inf)
        vout = numpy.hstack((vout, vnew))

    return vout


#-----
def random_augment(vin, k):
    """
  Add k random unit vectors to the stencil.

  This makes some of the theory work, but it is not always
  a good idea to do this. Adding more vectors makes stencil failure
  less likely, which is not good if you have a full stencil and you're
  wasting lots of time in the line search.

  Adding random vectors makes more sense if the minimizer is
  on a hidden or explicit constraint boundary.

  C. T. Kelley, January 5, 2011
  This code comes with no guarantee or warranty of any kind."""

    n = len(vin)
    if k == 0:
        vout = vin  # TODO: why? seems a dead assignment

    # Generate k uniformly distributed points on the unit sphere.
    # See Marsaglia, G. "Choosing a Point from the Surface of a Sphere."
    #     Ann. Math. Stat. 43, 645-646, 1972.
    rv = numpy.random.normal(0., 1., (n, k))
    for ir in range(k):
        rv[:,ir] /= numpy.linalg.norm(rv[:,ir])

    if rv.any():
        vout = numpy.hstack((vin, rv))

    return vout


#-----
def jac_or_grad(sgrad, jac, options):
    """
  Returns either the simplex gradient or Jacobian depending
  on the type of problem (optization or least squares).

  sdiff = jac_or_grad(sgrad, jac, options)"""

    least_squares = options.least_squares
    if least_squares == 1:
        sdiff = jac
    elif least_squares == 0:
        sdiff = sgrad
    else:
        raise ValueError('error in jac_or_grad')
    return sdiff



from __future__ import print_function
from ._linalg import eye, zeros, kk_proj, f_to_vals
from ._history import CompleteHistory, single_point_hist_update
import numpy
import operator
import warnings

__all__ = ['gauss_newton', 'qn_update']


#-----
def gauss_newton(f, x, fun, jac, xc,  gc, iteration_data, hessold):
    """
  Compute a damped Gauss-Newton step.
 
  xp, fvalp, funs, qfct, iarm, hess = \
          gauss_newton(f, x, fun, jac, xc, gc, iteration_data, hessold)
 
  You may elect to modify this routine if you don't want to spend the
  effort computing the Gauss-Newton model Hessian. The current version
  of imfil.m does not use it.
 
  Inputs: 
          f = objective function; f returns an residual vector.
          x = current point.
        fun = current (vector) residual at x.
        jac = stencil Jacobian at x.
         xc = previous point; xc is not used in this function, but
              that may change if we elect to put a quasi-Newton 
              method in here.
         gc = stencil gradient at xc.
  iteration_data = internal structure with many goodies inside; rtfm
    hessold = Gauss-Newton model Hessian at xc. Dummy argument
              waiting for quasi-Newton or Levenberg-Marquardt.
 
  Output:
       hess = Gauss-Newton model Hessian at x; no update done in here.
              It's in the argument lists only to enable a quasi-Newton
              update or a Levenberg-Marquardt iteration.
         xp = new point.
      fvalp = least squares error at xp = funs'*funs/2;
       funs = vector residual at xp.
       qfct = cost in function evaluations.
       iarm = number of step size reductions.
  diff_hist = history data for the Gauss-Newton loop
      nfail = 0 if the line search succeeds, 1 if it fails
         
  C. T. Kelley, January 10, 2011
  This code comes with no guarantee or warranty of any kind.
 
  Harvest what you need from iteration_data."""

    obounds   = iteration_data.obounds
    options   = iteration_data.options
    core_data = iteration_data.core_data
    h         = iteration_data.h

    # Compute stencil gradient (sgrad) and least squares error (fval)
    # at current point.
    funs = fun
    fval = fun.dot(fun)/2.
    sgrad = jac.dot(fun)

    hess = jac.dot(jac)  # Return the normal equations model Hessian for now

    n = len(x)

    # Get the epsilon-active indices and encode them in a diagonal matrix.
    epsb = 1.E-6;
    alist = zeros(len(x)).T
    for i in range(len(x)):
        alist[i] = (x[i] > obounds[i,0]+epsb) and (x[i] < obounds[i,1]-epsb)
    pr = numpy.diagflat(alist)

    # Compute the search direction with a QR factorization.
    rjac = jac.dot(pr)
    rq, rr = numpy.linalg.qr(rjac)
    sdir1 = (eye(n)-pr)*sgrad
    sdir2 = rq.dot(fun)
    sdir2 = numpy.linalg.lstsq(rr, sdir2.T, rcond=None)[0]
    sdir = sdir1+sdir2

    # Bound constrained line search
    qfct, xp, fvalp, iarm, fres, diff_hist, nfail = \
        armijo_explore(f, sdir, fval, x, h, core_data, obounds)

    # If the line search fails nothing changes.
    if len(fres) > 0:
       funs = fres

    return (xp, fvalp, funs, qfct, iarm, diff_hist, nfail, hess)


#-----
def qn_update(f, x, fval, sgrad, xc, gc, iteration_data, hessold):
    """
  hess, xp, fvalp, funs, qfct, iarm, diff_hist = \
          imfil_qn_update(f, x, fval, sgrad, xc, gc, iteration_data, hessold)
  
  Quasi-Newton update of point and Hessian. This function is never called
  for least squares problems.
 
  Input: 
         f = objective function.
         x = current point.
      fval = f(x). 
     sgrad = stencil gradient at current point x.
        xc = previous point.
        gc = stencil gradient at previous point xc.
  iteration_data = imfil internal structure
   hessold = previous model Hessian
 
  Output:
      hess = Quasi-Newton update Hessian at x.
        xp = new point.
     fvalp = f(xp)
     funs  = fvalp; Makes calling sequence compatible with Gauss-Newton
      qfct = cost in function evaluations.
      iarm = number of step size reductions.
  diff_hist= history data for the quasi-Newton loop
      nfail= 0 if line search succeeds, 1 if it fails
 
  The steps are (1) update Hessian to obtain hess(current Hessian),
                (2) use hess, gc, and the bounds to compute the new
                    search direction,
                (3) do a line search on that direction,
 
  C. T. Kelley, January 23, 2011
  This code comes with no guarantee or warranty of any kind."""

    core_data = iteration_data.core_data

    # obounds are the scaled 0-1 bounds. I use this field because my quasi-Newton
    # codes were written for general bounds and I do not want to invade them
    # any more than necessary.
    obounds = iteration_data.obounds
    h       = iteration_data.h

    # itc is the inner iteration counter. I am not updating the model Hessian
    # for the first inner iteration. The reason for this is that I need a 
    # couple gradients at each scale to make the quasi-Newton formula make sense.
    itc     = iteration_data.itc
    options = core_data.options
    quasi   = options.quasi

    nx = len(x)
    hess = hessold.copy()

    # Update the model Hessian for all but the initial inner iteration.
    if itc > 1:
        # Get the epsilon-inactive indices.
        epsb=1.E-6;
        alist = zeros(len(x)).T
        for i in range(len(x)):
            alist[i] = (x[i] > obounds[i,0]+epsb) and (x[i] < obounds[i,1]-epsb)

        if quasi == 1:        # BFGS
            hess = bfupdate(x, xc, sgrad, gc, hessold, alist)
        elif quasi == 2:      # SR1
            hess = sr1up(x, xc, sgrad, gc, hessold, alist)
        else:                 # nothing
            hess = eye(nx, nx)

    # Search direction
    sdir = numpy.linalg.lstsq(hess, sgrad, rcond=None)[0]

    # Bound constrained line search
    #
    qfct, xp, fvalp, iarm, fres, diff_hist, nfail = \
          armijo_explore(f, sdir, fval, x, h, core_data, obounds)
    funs = fvalp

    return (xp, fvalp, funs, qfct, iarm, diff_hist, nfail, hess)


#-----
def bfupdate(x, xc, sgrad, gc, hess, alist):
    """BFGS update of reduced Hessian, nothing fancy."""

    n = len(x)
    pr = numpy.diagflat(alist)
    y = sgrad-gc; s = x-xc; z = hess*s

    # Turn y into y#.
    y = pr*y
    if y.T*s > 0:
        hess = pr*hess*pr + (y*y.T/(y.T*s)) - pr*(z*z.T/(s.T*z))*pr
        hess = eye(n) - pr + hess

    if numpy.linalg.cond(hess) > 1.E6:
        hess = eye(n)

    return hess


#-----
def sr1up(x, xc, sgrad, gc, hess, alist):
    """SR1 update of reduced Hessian."""

    n = len(x)
    pr = numpy.diagflat(alist)
    y = sgrad-gc; s=x-xc
    z = y - hess*s

    # Turn y into y#.
    y = pr*y; z = pr*z
    if z.T*s != 0:
        ptst = z.T*(hess*z)+(z.T*z)*(z.T*z)/(z.T*s)
        if ptst > 0:
            hess = pr*hess*pr + (z*z.T)/(z.T*s);
            hess = eye(n) - pr + hess

    if numpy.linalg.cond(hess) > 1.E6:
        hess = eye(n)

    return hess


#-----
def armijo_explore(f, sdir, fold, xc, h, core_data, obounds):
    """
  Line search for imfil.m.
 
  C. T. Kelley, September 15, 2008
 
  This code comes with no guarantee or warranty of any kind.
 
  fct, x, fval, iarm, fres, diff_hist, nfail = \
        armijo_explore(f, sdir, fold, xc, h, core_data, obounds)
 
  This is an internal function, which you are NOT TO HACK! Since
  you may hack it anyway, I will tell you want is going on. If you
  break something, may Alberich's curse be upon you!
 
  Inputs:
          f = objective function.
       sdir = quasi-Newton search direction.
       fold = current function value.
   maxitarm = limit on number of step size reductions.
       beta = stepsize reduction factor.
          h = current scale.
  core_data = structure with the options + fun_data
              The options are documented in optset.m.
              fun_data is private to imfil.m and is used in the internal
              scaling for the function evaluation.

    obounds = Nx2 array of scaled bound constraints from imfil_core.

 Output:
        fct = cost of line search in function evaluations.
          x = new point.
       fval = f(x).
       iarm = number of step length reductions.  
       fres = residual for nonlinear least squares problems.
  diff_hist = history data for this iteration.
      nfail = 0 if the line search succeeds, 1 if it fails.

  Read the options array to get the ones we care about in the line search."""

    options = core_data.options

    if not options.parallel:
        fct, x, fval, iarm, aflag, fres, diff_hist, nfail = \
            serial_armijo(f, sdir, fold, xc, h, obounds, core_data)
    else:
        fct, x, fval, iarm, aflag, fres, diff_hist, nfail = \
            parallel_armijo(f, sdir, fold, xc, h, obounds, core_data)

    if iarm == options.maxitarm and aflag == 1 and options.verbose == 1:
        print('line search failure', iarm, h)

    return (fct, x, fval, iarm, fres, diff_hist, nfail)


#-----
def parallel_armijo(f, sdir, fold, xc, h, obounds, core_data):
    """
  Parallel line search for imfil.m.
  Uses extra processors to explore larger stencils.
 
  fct, x, fval, iarm, aflag, fres = \
     parallel_armijo(f, sdir, fold, xc, h, obounds, core_data)
 
  Before you modify this (to use a trust region method, say), consider
  writing your own with this as a template. I'll make this very easy
  in a later version of the code. 
 
  Inputs:
           f = objective function or residual vector, depending
               on whether this is a least squares problem or not.
        sdir = search direction.
        fold = current value of the objective.
        xc   = current point.
    maxitarm = limit on number of step size reductions.
        beta = reduction factor of step size. Currently beta = 1/2.
           h = current scale.
     obounds = Nx2 array of scaled bound constraints from imfil_core.
 
   core_data = structure with the options + fun_data
               The options are documented in optset.m.
               fun_data is private to imfil.m and is used in the internal
               scaling for the function evaluation.
  least_squares = least squares flag.
 
  Output:
         fct = cost in function evaluations.
           x = new point.
        fval = objective function value at new point = fres'*fres/2
               in the nonlinear least squares case.
       iarm  = number of step size reductions.
       aflag = 0 if the search finds a new point, = 1 if not.
       fres  = residual for nonlinear least squares problems
   diff_hist = history structure for this iteration
       nfail = 0 if the line search succeeds, 1 if it fails.
 
  C. T. Kelley, July 21, 2009
  This code comes with no guarantee or warranty of any kind."""

    options = core_data.options
    least_squares = options.least_squares
    beta          = options.armijo_reduction
    maxitarm      = options.maxitarm

    #Initialize the line search.
    lbda = 1
    n = len(xc)
    fct = 0
    aflag = 1
    dd = sdir
    fres = []
    frest = fres
    diff_hist = CompleteHistory()

    if options.limit_quasi_newton:
        # If the quasi-Newton step is much longer than the scale, shrink it.
        #
        # I am not convinced that this is the right thing to do, but it
        # really helps most of the time.
        #
        smax = 10*min(h, 1)
        if numpy.linalg.norm(dd) > smax:
            dd = smax*dd/numpy.linalg.norm(dd)

    x = xc
    fval = fold

    # Evaluate all steplength choices at once.
    number_steps = maxitarm+1
    ddm = zeros((n, number_steps))
    for i in range(number_steps):
        ddm[:,i] = xc-lbda*dd
        lbda = beta*lbda
        ddm[:,i] = kk_proj(ddm[:,i], obounds)

    fta, iflaga, ictra, tol = f(ddm, h, core_data)

    # Sort the points in failed and good into the history structure.
    for i in range(len(iflaga)):
        iflag = iflaga[i]
        if iflag:
            diff_hist.failed_points.append(ddm[:,i])
        else:
            diff_hist.good_points.append(ddm[:,i])
            if least_squares == 1:
                diff_hist.good_values.append(fta[:,i])
            else:
                diff_hist.good_values.append(fta[i])

    if least_squares:
        frest = fta

    fta = f_to_vals(fta, least_squares)
    fct += sum(ictra)
    ilose = (iflaga==1)
    fta[ilose] = fold+abs(fold)*.0001+1.E-12

    # Traditional duplicates a serial Armijo search. This will take
    # the longest possible step and is sometimes the right thing to do.
    #
    # Non-traditional will take the best point from all the
    # ones sampled in the search direction if the full step fails.
    #
    # I am still playing with this, and will let you chose in the options
    # command pretty soon. 
    traditional = 0
    if traditional == 0:
        it, ft = min(enumerate(fta), key=operator.itemgetter(1))
        xt = ddm[:, it]
        iarm = maxitarm
        if ft < fval:
            aflag = 0
            fval = ft
            if least_squares:
                fres = frest[:,it]
            x = xt
            iarm = min(it, maxitarm-1)
        else:
            iarm -= 1

            while iarm < maxitarm and aflag == 1:
                ft = fta[iarm+2]
                xt = ddm[:,iarm+2]
                if ft < fval and aflag == 1:
                    aflag = 0
                    if least_squares:
                        fres = frest[:,iarm+2]
                    fval = ft
                    x = xt
                iarm += 1

    if iarm == maxitarm and aflag == 1 and options.verbose == 1:
        print('line search failure', iarm, h)

    nfail = aflag
    return (fct, x, fval, iarm, aflag, fres, diff_hist, nfail)


#-----
def serial_armijo(f, sdir, fold, xc, h, obounds, core_data):
    """
  Serial line search for imfil.m.
 
  fct, x, fval, iarm, aflag, fres = \
          serial_armijo(f, sdir, fold, xc, h, obounds, core_data)
 
  This is a plain vanilla serial line search. Nothing to see here;
  move along.
 
  Before you modify this (to use a trust region method, say), consider
  writing your own with this as a template. I'll make this very easy
  in a later version of the code.
 
  Inputs:
            f = objective function or residual vector, depending
                on whether this is a least squares problem or not.
         sdir = search direction.
         fold = current value of the objective.
           xc = current point.
            h = current scale.
      obounds = Nx2 array of scaled bound constraints from imfil_core.
 
    core_data = structure with the options + fun_data
                The options are documented in optset.m.
                fun_data is private to imfil.m and is used in the internal
                scaling for the function evaluation.
 
  Output:
         fct = cost in function evaluations.
           x = new point.
        fval = objective function value at new point = fres'*fres/2
               in the nonlinear least squares case.
       iarm  = number of step size reductions.
       aflag = 0 if the search finds a new point, = 1 if not.
       fres  = residual for nonlinear least squares problems
   diff_hist = history structure for this iteration
       nfail = 0 if the line search succeeds, 1 if it fails.

  C. T. Kelley, September 15, 2008
  This code comes with no guarantee or warranty of any kind."""

    options = core_data.options;
    least_squares = options.least_squares
    beta          = options.armijo_reduction
    maxitarm      = options.maxitarm

    # Initialize the line search.
    lbda = 1
    n = len(xc)
    iarm = -1
    fct = 0
    aflag = 1
    dd = sdir
    fres = []
    frest = []
    diff_hist = CompleteHistory()

    if options.limit_quasi_newton:
        # If the quasi-Newton step is much longer than the scale, shrink it.
        #
        # I am not convinced that this is the right thing to do, but it
        # really helps most of the time.
        smax = 10*min(h, 1)
        if numpy.linalg.norm(dd) > smax:
            dd = smax*dd/numpy.linalg.norm(dd)

    x = xc
    fval = fold
    while iarm < maxitarm and aflag == 1:
        d = -lbda*dd
        xt = x+d
        xt = kk_proj(xt, obounds)
        try:
            ft, ifl, ict, tol = f(xt, h, core_data)
        except Exception as e:
            warnings.warn('dropping function evaluation "%s"' % (str(e),))
            continue

        fct += ict
        single_point_hist_update(diff_hist, xt, ft, ifl)
        if least_squares == 1:
            frest = ft
            ft = numpy.norm(frest)/2.
        if ifl == 1:
            ft = fold+abs(fold)*.0001+1.E-12
        if ft < fval and aflag == 1:
            aflag, fval, fres, x = 0, ft, frest, xt
        if aflag == 1:
            lbda = beta*lbda
        iarm += 1

    if iarm == maxitarm and aflag == 1 and options.verbose == 1:
        print('line search failure', iarm, h)

    nfail = aflag
    return (fct, x, fval, iarm, aflag, fres, diff_hist, nfail)



from __future__ import print_function
import numpy

__all__ = ['CompleteHistory', 'append_history', 'single_point_hist_update',
    'many_point_hist_update', 'scan_history', 'scan_complete_history']


#-----
class CompleteHistory(object):
    def __init__(self, good_points=[], good_values=[], failed_points=[]):
        self.good_points   = list(good_points)
        self.good_values   = list(good_values)
        self.failed_points = list(failed_points)

    def copy(self):
        return CompleteHistory(
            self.good_points[:], self.good_values[:], self.failed_points[:])

    def __str__(self):
        return "good_points = %s\n good_values = %s\n failed_points = %s\n" %\
            (str(self.good_points), str(self.good_values), str(self.failed_points))


#-----
def append_history(histout, fcount, fval, npgrad, stepn, iarm, x):
    try:
        if len(fval) == 1:
            fvalhist = float(fval)
    except TypeError:
        fvalhist = fval
    histout.append([fcount, fvalhist, npgrad, stepn, iarm] + list(map(float, x)))


#-----
def single_point_hist_update(hist, x, fout, ifail):
    """
  Update the complete_history structure after a single call to f.
 
  single_point_hist_update(hist, x, fout, ifail)"""

    # Write the data.
    if ifail == 1:
        hist.failed_points.append(x)
    else:
        hist.good_points.append(x)
        hist.good_values.append(fout)


#-----
def many_point_hist_update(hist, diff_hist, inplace=True):
    """
  many_point_hist_update(hist, diff_hist)
 
  Update the complete_history structure after a many calls to f.
  
  WARNING! The complete_history structure uses your bounds, and is
  not scaled to make 0 <= x(i) <= 1."""

    if not inplace:
        hist = hist.copy()
    if type(diff_hist.good_points) == list:
        hist.good_points.extend(diff_hist.good_points)
    else:
        for i in range(diff_hist.good_points.shape[1]):
            hist.good_points.append(diff_hist.good_points[:,i])
    hist.good_values.extend(diff_hist.good_values)
    hist.failed_points.extend(diff_hist.failed_points)
    return hist


#-----
def scan_history(complete_history, xp, fp, dx):
    """
  Find previous evaluations, if any, at the requested points <xp>."""

    oldresults = {}
    newpoints = {}
    for i in range(xp.shape[1]):
        point = xp[:,i]
        try:
            fpt, ift = scan_complete_history(complete_history, point)
            oldresults[i] = (fpt, ift)
        except LookupError:
            newpoints[i] = point

    return oldresults, newpoints


#-----
def scan_complete_history(complete_history, x):
    """
  Search for a previously evaluated point within 1.E-12 of <x>,
  returns previous value and associated flag."""

    for i, point in enumerate(complete_history.good_points):
        d = numpy.linalg.norm(x-point, ord=numpy.inf)
        if d < 1.E-12:
            return complete_history.good_values[:,i], 0

    for i, point in enumerate(complete_history.failed_points):
        d = numpy.linalg.norm(x-point, ord=numpy.inf)
        if d < 1.E-12:
            return numpy.NaN, 1

    raise LookupError("no such point")



from __future__ import print_function
import copy, string

__all__ = ['optset']

STANDALONE = True

def optset(optin=None, **optkwds):
    """ 
  Set the options in imfil.
  C. T. Kelley, Nov 11, 2007
 
  optout = optset(name1=val1, name2=val2, ...)
           creates the options dictionary with the named options
           set to something other than the defaults.
 
  optout = optset(optin, name1=val1, name2=val2, ...);
           modifies the dictionary <optin> and creates a new structure
           <optout>.
 
  And now for the lists. The details of the options, the motiviation
  for the defaults, and suggestions for their use live in the manual.
 
  Scalar Options:
 
  armijo_reduction: Reduction factor for the linesearch. The default is .5.
 
  complete_history: 1 or 'on' to store complete history of all calls to f. 0
         or 'off' otherwise. The default is 'off'. Do not turn this off unless
         you are having severe problems with storage.

  fscale: Sets a "typical value of f", used to improve the quality of the
         difference gradient. The default is imfil_fscale = 1.2*| f(x_0) |.
         Setting options = optset(options, fscale=-c); with c > 0 will set
         imfil_fscale = c * |f(x_0)| and options = optset(options, fscale=c)
         with c > 0 will set imfil_fscale = c.

  function_delta: Set this to an estimate for the absolute error in f if you
         have one. Ignore this option if you do have no idea how accurate your
         function evaluation is. Set function_delta > 0 and the optimization
         will terminate if two successful iterations (ie. iterations that
         change the best point) have function values that differ
         by < function_delta. The default is -1, which means that
         function_delta does not affect the iteration.

  stencil_delta: Set this to an estimate for the absolute error in f if you
         have one. Ignore this option if you do have no idea how accurate your
         function evaluation is. The optimization loop will terminate when the
         maximum absolute difference of function values on the stencil
         is < stencil_delta. The default is -1, which means that stencil_delta
         has no effect.

  least_squares: imfil does the smart thing if you're solving a nonlinear
         overdetermined (m >=n) least squares problem. Set least_squares=1 and
         let your function return the vector residual in R^m. imfil will
         compute f'*f/2 as part of the optimization.

  limit_quasi_newton: 'yes' or 1 means that the quasi-Newton direction will be
         truncated to a length of at most 10*h, where h is the current scale.
         This is a good idea if the function is very noisy, as it can keep the
         line search from thrashing about. The code is different from the old
         FORTRAN version, and now I am not sure about what the best value is.
         'yes' is currently the default. 'no' or 0 means the quasi-Newton
         direction vector gets left alone.

  maxit: The quasi-Newton inner iteration is limited to maxit*n iterations
         for each scale. The default is 50.

  maxitarm: The line search will reduce the step by at most maxitarm times
         before declaring failure. The default is 3. The line search is on a
         short leash for good reason. If you don't find something useful after
         three reductions, you're not likely to do better with more effort.

  maxfail: After maxfail line search or stencil failures, imfil stops
         the entire iteration. The default is maxfail = 3.

  noise_aware: Set this to one if your function can return as estimate of the
         norm of the noise. The call is [fout,ifail,icount,noise_val]=f(x).
         In this case imfil will declare stencil failure if the difference
         between the max and min values on the stencil is < noise_val.
         The default is 0 (not noise_aware).
         You can set scale_aware=1 and noise_aware=1 at the same time.

         noise_val is not the same as stencil_delta. You set noise_val via the
         call to f as f's best estimate of the error. noise_val is used as a
         test to reject the stencil and reduce the scale.
         stencil_delta is used to terminate the optimization.

  parallel: Set this to one if f can accept multiple input vectors and
         exploit parallelism. The output should be vectors of
         [fout,ifail,icount], with a length equal to the number of input
         variables. When imfil computes the stencil gradient f will get all
         the directions at once. It is then your job to manage the parallel
         computation for optimal load balancing.

  quasi: Your choices for quasi-Newton acceleration are none (0),
         BFGS ('bfgs'), or SR-1 ('sr1'). The default as of today is BFGS, but
         I may change that as I have done in the past.

  random_stencil: Set this to k > 0 and k random unit vectors will be
         added to the stencil before each call to the simplex gradient.
         The default is 0.

  scale_step, _start, _depth: The range of scales is scale_step^-scale_start
         to scale_step^-scale_depth. The defaults are scale_step=2,
         scale_start=3, scale_depth=11.

  scale_aware: Setting this to 1 tells imfil that f can do something smart if
         it knows the scale. The call then becomes [fout,ifail,icount]=f(x,h)
         The scale is the second input argument. The default is 0 (not
         scale-aware). You can set scale_aware=1 and noise_aware=1 at the same
         time.

  simple_function: Set this to 1 and imfil will not ask you for either a cost
         estimate or a failure flag. Instead imfil will assume all costs are
         the same (and set them to 1) and that your function never, EVER,
         fails. Your function can then return only the value (or values if
         parallel=1). So the calling sequence is

            fout=f(x,h) (where h is optional, see scale_aware)

         The default is 0 (your function returns three things).

  smooth_problem: Set this to 1 and you will configure imfil to solve smooth
         problems. This option is equivalent to

                  bscales=[.5, .01, .001, .0001, .00001];
                    options=optset(
                        custom_scales=bscales,
                        stencil_wins='yes',
                        limit_quasi_newton='off',
                        armijo_reduction=0.25,
                        maxitarm=5)

         If you use this option you should probably increase your budget and
         experiment with both the default BFGS quasi-Newton method and SR1.

  stencil: imfil has three built-in stencils, with room for you to create you
        own in the matrix options. Setting stencil lets you chose the centered
        difference stencil (0), the forward difference stencil (1), and the
        positive basis stencil (2). The default is the centered difference
        stencil (stencil = 0). Choosing the forward difference stencil will
        cause a warning, as that option will go away in a future version.

  stencil_wins: Setting this to 'yes' or 1 will make the new point the better
        of the best point in the stencil and the point returned by the line
        search. The default 'no' is a bias in favor of the quasi-Newton point,
        and sets the point to the quasi-Newton point if the line search
        succeeds, and to the best point in the stencil only if the line search
        fails.

  target: You may have either a good lower estimate of the optimal value of f,
        or an idea about when the optimization has done enough. If you do, set
        target to that value and the optimization will stop when f < target.

  termtol: The inner iteration will terminate successfully if
        fscale*norm(proj(difference_grad)) < termtol * h.
        The default is termtol = .01.
        The inner iteration usually terminates with stencil failure, so there
        is little reason to change termtol unless you have a smooth problem,
        and then you might want to reduce it. Increasing termtol will reduce
        the number of total calls to f, but may also lead to premature
        termination and very little progress.

  svarmin: Set svarmin if you have a good estimate of the noise.
        If svarmin > 0 then imfil will declare stencil failure if the
        difference between the max and min values on the stencil is < svarmin.
        This option is the same as setting noise_aware to 1 and letting the
        function return svarmin as the noise_level. If the noise is not a
        function of h, setting svarmin once is easier.


  Matrix options

  custom_scales: You may set you own scales with an array
        H=[h(1), ... h(smax)] where 1 > h(1) > .. > h(smax) > 0
        The syntax is optout=optset(optin, custom_scales=H)

  vstencil: You may replace the built-in stencils with your own custom
        design. If VS is your N x M list of M directions, then
            optout=optset(optin, vstencil=VS)
        will get it done.

  Advanced Options: WARNING!!! You have to be aware of imfil_core's internal
  scaling to use this stuff. RTFM. I am not able to help you with your use of
  these options. You are on your own. Please read the documentation and look
  at the examples.

  add_new_directions: 
         This function, which you must provide, will examine the current
         stencil, scale, and history of the iteration and add directions to
         the stencil. You may want to do this, for example, to add tangent
         directions for nearly active linear constraints.

         You enable this with
           options=optset(options, add_new_directions=mystencil);
         where the calling sequence of your function must be

           Vadd=mystencil(x,h,V);

           where x = current iterate;
                 h = current scale;
                 V = stencil for this iteration (which may include any random
                     directions you asked for with the random_stencil option)

  explore_function:
         This function, which you write, will explore design space after each
         inner iteration is complete.

         Turn it on with
           options=optset(options, explore_function=my_search);
         where the calling sequence of your search function must be

           [xs, fs, complete_history] = my_search(complete_history)

         xs is the best point from your search, fs=f(xs), and you must update
         the complete_history array, so that array is passed back and forth to
         your function.

         Setting this option also sets explore to 1

  executive_function:
         You write an executive function to replace the internal quasi-Newton
         or Gauss-Newton solvers in imfil.

         Turn the executive_function on with
           options=optset(executive_function=my_solver,
                          executive_data=my_data)
         Your function is my_solver and the data you function needs is
         my_data. You must include my_data in the call even if you don't use
         it.

         The calling sequence for your function is
           [xp, fvalp, funp, fcost, iarm, solver_hist, nfail, new_data]
             = my_solver(f, x, fun, sdiff, xc,  gc, iteration_data, old_data)

         Please, PLEASE, read the manual before messing with the executive
         function.


  ----------------------------------------------------------------

  The extra_argument options are set for you when you add the extra
  argument to your call to imfil. There is no reason to set them
  with imfil_optset.

  extra_argument: 1 or 'yes' if you are passing data through imfil 
                  to the function, 
                  0 or 'no' other wise

  extra_arg_value: the data you are passing to the function if 
                   'extra_argument' is on.

  ----------------------------------------------------------------

  Please look at the example for this in the software collection and in
  the users' guide."""

    if not optin:
    # Start from default structure
        optout = _create_defaults()
    else:
        optout = copy.copy(optin)

    for name, val in optkwds.items():
        optout = _optset_one(name, val, optout)

    return optout


#-----
def _create_defaults():
    """Default values"""

    class OptSet(object):
        __slots__ = {
            'add_new_directions' : [],  # function to add new directions.
            #
            'armijo_reduction' : .5,    # Step size reduction factor for Armijo rule.
            #
            'complete_history' : 0,     # Store complete history.
            #
            'custom_scales' : [],       # vector of custom scales
            #
            'executive' : 0,            # No executive.
            #
            'executive_function' : [],  # What function?
            #
            'executive_data' : [],      # What data?
            #
            'executive_data_flag' : 0,  # No data! Really!
            #
            'explore_data_flag' : 0,    # The exploration step needs no data.
            #
            'explore_data' : [],        # No data means no data.
            #
            'explore' : 0,              # no exploration step
            #
            'explore_function' : [],    # If I'm not exploring, why write a function?
            #
            'extra_argument' : 0,       # passing something to f?
            #
            'extra_arg_value' : [],     # what you pass to f if extra_argument = 1
            #
            'fscale' : -1.2,            # Scale function by 1.2 times abs(f(x_0)).
            #
            'function_delta' : -1,      # Do not terminate on small differences of successive
                                        # successful iterations.
            #
            'stencil_delta' : -1,       # Do not terminate the optimization 
                                        # on small total variation of f.
            #
            'least_squares' : 0,        # Is this a nonlinear least squares problem?
            #
            'limit_quasi_newton' : 1,   # Don't let the quasi-Newton direction get too long.
            #
            'maxit' : 50,               # At most maxit*n quasi-Newton iterations/scale.
            #
            'maxitarm' : 3,             # Limit on number of step length reductions.
            #
            'maxfail' : 3,              # Limit on consecutive line search/stencil failures.
            #
            'noise_aware' : 0,          # Can your function estimate the noise? Default = no.
            #
            'parallel' : 0,             # Serial is the default.
            #
            'quasi' : 1,                # BFGS
            #
            'random_stencil' : 0,       # Leave the stencil alone.
            #
            'scale_aware' : 0,          # Can your function use the scale?
            #
            'scale_step' : 2,           # Stepping factor for scales
            #
            'scale_start' : 1,          # Scales begin at h=(1/scale_step)^scale_start
            #
            'scale_depth' : 8,         # Scales end at h=(1/scale_step)^(scale_start+scale_depth)
            #
            'simple_function' : 0,      # Your function returns more than just the value.
            #
            'smooth_problem' : 0,       # Don't use imfil for smooth problems if you can
                                        # avoid it.
            #
            'standalone': STANDALONE,   # Use orginal interface if standalone, scikit-quant iface if not
            #
            'stencil' : 0,              # Central difference stencil.
            #
            'stencil_wins' : 0,         # Who's the new point?
            #
            'target' : -1.E8,           # Target for optimal value = -1.d8.
                                        # You should try to do better.
            #
            'termtol' : .01,            # Small enough to let stencil failure do its job.
            #
            'svarmin' : 0,              # I have no idea what the noise is. 
            #
            'verbose' : 0,              # Shut up.
            #
            'vstencil' : [],            # No fancy stencil.
        }

        def __init__(self):
            for key, value in self.__slots__.items():
                setattr(self, key, value)

    return OptSet()


#-----
def _optset_one(name, val, optin):
    """
  This is the internal function to manage the options structure."""

    name = name.lower();

    # Is this an advanced option?
    if name in ('add_new_directions', 'extra_arg_value',
                'explore_function', 'executive_function'):
        adv_flag = 1
    else:
        adv_flag = 0

    # Test for verbal shortcuts and do the right thing.
    # The advanced options are set to names of functions and 
    # do not use verbal shortcuts.
    if isinstance(val, str) and adv_flag == 0:
        optin = verbal_shortcut(name, val, optin)
    else:
        # Update options.
        setattr(optin, name, val)

    # Now put in the flags for the explore_function and executive options.
    if name == 'explore_function':
        optin.explore = 1
    elif name == 'explore_data':
        optin.explore_data_flag = 1
    elif name == 'explore':
        yesno = optin.explore
        if not yesno:
            optin.explore_function  = []
            optin.explore_data_flag = 0
            optinexplore_data       = []
    elif name == 'executive_function':
        optin.executive = 1
    elif name == 'executive_data':
        optin.executive_data_flag = 1
    elif name == 'executive':
        yesno = getfield(optin, 'explore');
        if not yesno:
            optin.executive_function  = []
            optin.executive_data_flag = 0
            optin.executive_data      = []

    return optin


#-----
def verbal_shortcut(name, val, optin):
    """
  Plugs into imfil_optset and lets you use verbal shortcuts
  to control options with yes/no, on/off, or short lists of 
  options."""

    val = val.lower();
    if name == 'add_new_directions':
       return optin
    elif name == 'complete_history':
        valout = yes_no(val)
    elif name == 'explore':
        valout = yes_no(val)
    elif name == 'least_squares':
        valout = yes_no(val)
    elif name == 'limit_quasi_newton':
        valout = yes_no(val)
    elif name == 'quasi':
        if val == 'bfgs':
            valout = 1
        elif val == 'sr1':
            valout = 2
        elif val == 'none':
            valout = 0
        else:
            raise ValueError('unexpected value for %s:' % name, val)
    elif name == 'scale_aware':
        valout = yes_no(val)
    elif name == 'simple_function':
        valout = yes_no(val)
    elif name == 'smooth_problem':
        valout = yes_no(val)
    elif name == 'stencil_wins':
        valout = yes_no(val)
    elif name == 'verbose':
        valout = yes_no(val)
    else:
        raise ValueError('no verbal shortcut found for %s' % str(name))

    setattr(optin, name, valout)
    return optin


#-----
def yes_no(val):
    if val in ['yes', 'on']:
        return True
    if val in ['no', 'off']:
        return False
    raise ValueError(val)



import py, os, sys
import numpy as np

sys.path = [os.path.join(os.pardir, 'python')] + sys.path


class TestIMFIL:
    def setup_class(cls):
        import SQImFil, logging
        SQImFil.log.setLevel(logging.DEBUG)

    def test01_simple_example(self):
        """Read access to instance public data and verify values"""

        import SQImFil

        def f_easy_simple(x):
            from math import sin

            fv = np.inner(x, x)
            fv *= 1 + 0.1*sin(10*(x[0]+x[1]))

            return fv

        def f_easy(x):
            from math import sin

            fv = np.inner(x, x)
            fv *= 1 + 0.1*sin(10*(x[0]+x[1]))

            return (fv, 0, 1)

        def f_easy_parallel(ax):
            from math import sin

            if len(ax.shape) == 1:
                return f_easy(ax)

            res = list()
            for i in range(ax.shape[0]):
                res.append(f_easy(ax[i])[0])
            return (res, [0]*len(res), [1]*len(res))

        bounds = np.array([[-1, 1], [-1, 1]], dtype=float)
        budget = 40
        x0 = np.array([0.5, 0.5])

        opt_common = {'scale_depth' : 7, 'complete_history' : True}
        for func in [f_easy_simple, f_easy]:#, f_easy_parallel]:
            if func == f_easy:
                optset = SQImFil.optset(simple_function=False, **opt_common)
            elif func == f_easy_parallel:
                optset = SQImFil.optset(parallel=True, **opt_common)
            else:
                optset = SQImFil.optset(**opt_common)
            res, histout, complete_history = SQImFil.minimize(func, x0, bounds, budget, optset)
            assert type(res.optpar) == np.ndarray
            # this problem is symmetric, so values may have switched; for
            # simplicity, just check the sum
            assert np.round(sum(res.optpar)-sum((-0.00681757, 0.00880743)), 8) == 0
            assert len(histout) == 10
            assert np.round(histout[9,5:7].sum()-sum((-6.81757191e-03, 8.80742809e-03)), 8) == 0

        # TODO: figure out why the parallel version performs a bit better
        optset = SQImFil.optset(parallel=True, **opt_common)
        res, histout, complete_history = \
             SQImFil.minimize(f_easy_parallel, x0, bounds, budget, optset)
        assert type(res.optpar) == np.ndarray
        assert np.round(sum(res.optpar)-sum((0.00281554, 0.00281554)), 5) == 0

    def test02_simple_example_multi_local(self):
        """Verify with several local minima (ie. start point dependent"""

        import SQImFil

        def f_easy(x):
            from math import sin, cos, pi

            fv = x.T.dot(x)
            fv *= 1 + 0.1*sin(10*(x[0]+x[1])) - cos(pi*x[2]/2.)

            return (fv, 0, 1)

        import numpy as np
        bounds = np.array([[-1, 1], [-1, 1], [0, 1]], dtype=float)
        budget = 40
        optset = SQImFil.optset(scale_depth=7, complete_history=0)

        x0 = np.array([0., 0., 0.])
        res, histout = SQImFil.minimize(f_easy, x0, bounds, budget, optset)
        assert np.round(res.optval, 5) == -0.12422

        x0 = np.array([-1.0, -1.0, 0.011198])
        res, histout = SQImFil.minimize(f_easy, x0, bounds, budget, optset)
        assert np.round(res.optval, 5) == -0.18258

        x0 = np.array([-1.0, 0.222183, 0.037152])
        res, histout = SQImFil.minimize(f_easy, x0, bounds, budget, optset)
        assert np.round(res.optval, 5) == -0.10459

        x0 = np.array([-0.8305, 0., 0.314])
        res, histout = SQImFil.minimize(f_easy, x0, bounds, budget, optset)
        assert np.round(res.optval, 5) == -0.18259

        xmin = np.array([-1.0, -1.0, 0.0])
        assert type(res.optpar) == np.ndarray
        assert f_easy(res.optpar)[0] <= f_easy(x0)[0]
        assert res.optval            <= f_easy(x0)[0]
        assert np.round(sum(res.optpar-xmin), 8) == 0
        assert f_easy(res.optpar)[0] == f_easy(xmin)[0]
        assert round(res.optval-f_easy(res.optpar)[0], 8) == 0



from __future__ import print_function
import numpy

__all__ = ['error_check', 'check_bounds', 'check_first_eval', 'isok']


#-----
def isok(x, bounds):
    # test whether all values of x fall within their bounds
    il = (x >= bounds[:, 0]).all()
    iu = (x <= bounds[:, 1]).all()
    return il and iu


#-----
def error_check(*varargin):
    """
  Checks dimensions and options for input errors or inconsistencies.
  This function is not intended for use outside of imfil.m. It does
  not error check itself.
 
  C. T. Kelley, Feb 19, 2010"""
 
    name = varargin[0]
    if name == 'bounds':
        x = varargin[1]
        bounds = varargin[2]
        return check_bounds(x, bounds)
    elif name == 'first_eval':
        funs = varargin[1]
        options = varargin[2]
        iflag = varargin[3]
        return check_first_eval(funs, options, iflag)

    raise ValueError('"%s" is an illegal argument to error_check.' % name)


#-----
def check_bounds(x, bounds):
    """
  Sanity check for the bounds. Complain if the bounds are not the 
  same length as x, if the bounds are not properly arranged into 
  columns, if the lower bound is not less than the upper bound,
  or if x is infeasible."""
 
    if x.shape[0] != bounds.shape[0]:
        raise ValueError('The columns [lower, upper] of the bounds array must '
                         'have the same length as x. ')

    if bounds.shape[1] != 2:
        raise ValueError('The bounds array must have two columns [lower, upper].')

    diff_vec = bounds[:,1] - bounds[:,0]
    m_vec = min(diff_vec)
    if m_vec <= 0:
        raise ValueError('The lower bound must be strictly greater than the upper bound.')

    if m_vec > 0:
        px = numpy.maximum(bounds[:,0], numpy.minimum(x, bounds[:,1]))
        nd = numpy.linalg.norm(x-px)
        if nd > 0:
            raise ValueError('The initial iterate is infeasible. '
                             'x must satisfy the bound constraints.')


#-----
def check_first_eval(funs, options, iflag):
    """
  Complain if you have a scalar-valued function + nonlinear least squares.
  Give up if you have a vector-valued function + not a least squares problem.
  Warn if the function fails to return a value at the initial iterate."""

    lsqerr = 0
    try:
        mf, nf = funs.shape
    except AttributeError:
        mf = nf = 1
    lsq = options.least_squares
    if mf > 1 and lsq == 0:
        print('Your function is vector-valued but the least_squares option is off.')
        print('imfil cannot run with this inconsistency.')
        lsqerr = 1

    elif mf == 1 and lsq == 1:
        print('Your function is scalar-valued but the least_squares option is on.')
        print('Are you sure this is what you want to do?')

    if iflag > 0:
        print('Your function failed to return a value at the initial iterate.')
        print('This is usually a problem. You have been warned.')

    return lsqerr



from __future__ import print_function

### temporary override of NumPy deprecation warning, see:
### https://docs.scipy.org/doc/numpy/user/numpy-for-matlab-users.html
import warnings
warnings.filterwarnings('ignore', category=PendingDeprecationWarning)

import logging

log = logging.getLogger('SKQ.ImFil')

log.info("""
------------------------------------------------------------------------
C.T. Kelley, "Implicit Filtering", 2011, ISBN: 978-1-61197-189-7
Software available at ctk.math.ncsu.edu/imfil.html
------------------------------------------------------------------------""")

__all__ = ['minimize', 'optset', 'log']

from ._imfil import minimize
from ._optset import optset
